/* This is the header file of the Parma Watchdog Library.
   Copyright (C) 2002-2003 Roberto Bagnara <bagnara@cs.unipr.it>

This file is part of the Parma Watchdog Library (PWL).

The PWL is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

The PWL is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

For the most up-to-date information see the CS@Parma software
site: http://www.cs.unipr.it/Software/ . */

#ifndef PWL_pwl_hh
#define PWL_pwl_hh 1

#ifdef NDEBUG
#define PWL_SAVE_NDEBUG 1
#undef NDEBUG
#endif

/* config.h.  Generated by configure.  */
/* config.h.in.  Generated from configure.ac by autoheader.  */

/* Define to 1 if you have the <dlfcn.h> header file. */
#define HAVE_DLFCN_H 1

/* Define to 1 if you have the <inttypes.h> header file. */
#define HAVE_INTTYPES_H 1

/* Define to 1 if you have the <memory.h> header file. */
#define HAVE_MEMORY_H 1

/* Define to 1 if you have the `setitimer' function. */
#define HAVE_SETITIMER 1

/* Define to 1 if you have the <stdint.h> header file. */
#define HAVE_STDINT_H 1

/* Define to 1 if you have the <stdlib.h> header file. */
#define HAVE_STDLIB_H 1

/* Define to 1 if you have the <strings.h> header file. */
#define HAVE_STRINGS_H 1

/* Define to 1 if you have the <string.h> header file. */
#define HAVE_STRING_H 1

/* Define to 1 if you have the <sys/stat.h> header file. */
#define HAVE_SYS_STAT_H 1

/* Define to 1 if you have the <sys/time.h> header file. */
#define HAVE_SYS_TIME_H 1

/* Define to 1 if you have the <sys/types.h> header file. */
#define HAVE_SYS_TYPES_H 1

/* Define to 1 if you have the <unistd.h> header file. */
#define HAVE_UNISTD_H 1

/* Assertions are disabled when this is defined */
#define NDEBUG 1

/* Name of package */
#define PACKAGE "watchdog"

/* Define to the address where bug reports for this package should be sent. */
#define PACKAGE_BUGREPORT "bagnara@cs.unipr.it"

/* Define to the full name of this package. */
#define PACKAGE_NAME "the Parma Watchdog Library"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "the Parma Watchdog Library 0.2"

/* Define to the one symbol short name of this package. */
#define PACKAGE_TARNAME "watchdog"

/* Define to the version of this package. */
#define PACKAGE_VERSION "0.2"

/* Define to 1 if you have the ANSI C header files. */
#define STDC_HEADERS 1

/* Define to 1 if you can safely include both <sys/time.h> and <time.h>. */
#define TIME_WITH_SYS_TIME 1

/* Version number of package */
#define VERSION "0.2"

/* Define to empty if `const' does not conform to ANSI C. */
/* #undef const */

#if __GNUC__ < 3
// Get around a bug in GCC versions prior to 3 whereby the <cassert>
// header is wrongly guarded against multiple inclusion.
#ifdef __CASSERT__
#undef __CASSERT__
#endif

#endif

#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
#endif

#include <cassert>
#include <list>
#include <stdexcept>
namespace Parma_Watchdog_Library {
  class Watchdog;
  class Flag;
  class Init;
}
namespace Parma_Watchdog_Library {


extern "C" void PWL_handle_timeout(int signum);


class Watchdog {
public:
  template <typename Flag_Base, typename Flag>
  Watchdog(int units, const Flag_Base* volatile* holder, Flag& flag);

  Watchdog(int units, void (*function)());
  ~Watchdog();


private:
  friend class Init;
  static void initialize();
  static void finalize();


  class Time {
  private:
    int secs;
    int microsecs;

  public:

    Time();

    explicit Time(unsigned int units);

    void set(int s, int m);

    void reset();

    int seconds() const;

    int microseconds() const;

    Time& operator +=(const Time& y);

    Time& operator -=(const Time& y);

    Time operator+(const Time& y) const;
    Time operator-(const Time& y) const;
    bool operator==(const Time& y) const;
    bool operator!=(const Time& y) const;
    bool operator<(const Time& y) const;
    bool operator<=(const Time& y) const;
    bool operator>(const Time& y) const;
    bool operator>=(const Time& y) const;
  };

  class Pending_Element;
  friend class Pending_Element;
  friend Time& Time::operator+=(const Time& y);
  friend Time& Time::operator-=(const Time& y);
  friend Time Time::operator+(const Time& y) const;
  friend Time Time::operator-(const Time& y) const;


  class Handler {
  public:
    virtual void act() const = 0;
  };

  template <typename Flag_Base, typename Flag>
  class Handler_Flag : virtual public Handler {
  private:
    const Flag_Base* volatile* holder;
    Flag& flag;

  public:
    Handler_Flag(const Flag_Base* volatile* h, Flag& f);
    void act() const;
  };

  class Handler_Function : virtual public Handler {
  public:
    Handler_Function(void (*f)());
    void act() const;

  private:
    void (*function)();
  };


  class Pending_Element {
  public:
    Time deadline;
    const Handler* handler;
    bool* p_expired_flag;
    Pending_Element(const Time& d, const Handler* h, bool* p);
  };

  typedef std::list<Pending_Element> Pending;

  bool expired;
  const Handler* handler;
  Pending::iterator pending_position;

private:

  Watchdog(const Watchdog&);
  Watchdog& operator=(const Watchdog&);


  static itimerval current_timer_status;


  static void get_timer(Time& time);


  static itimerval signal_once;


  static Time last_time_requested;


  static void set_timer(const Time& time);


  static void stop_timer();


  static void reschedule();


  static Time reschedule_time;


  static Time time_so_far;


  static Pending pending;


  static void handle_timeout(int);


  static Pending::iterator insert_pending(const Time& deadline,
                                          const Handler* handler,
                                          bool* p_expired);


  static Pending::iterator new_watchdog_event(int units,
                                              const Handler* handler,
                                              bool* p_expired);


  void remove_watchdog_event(Pending::iterator position);


  static volatile bool alarm_clock_running;


  static volatile bool in_critical_section;

  friend void Parma_Watchdog_Library::PWL_handle_timeout(int signum);
};

class Init {
private:

  static unsigned int count;

public:

  Init();


  ~Init();
};

}


namespace {

Parma_Watchdog_Library::Init Parma_Watchdog_Library_initializer;

}
namespace Parma_Watchdog_Library {

template <typename Flag_Base, typename Flag>
Watchdog::Handler_Flag<Flag_Base,
                       Flag>::Handler_Flag(const Flag_Base* volatile* h,
                                           Flag& f)
                         : holder(h), flag(f) {
}

template <typename Flag_Base, typename Flag>
void
Watchdog::Handler_Flag<Flag_Base, Flag>::act() const {
  if (*holder == 0
      || static_cast<const Flag*>(*holder)->priority() < flag.priority())
    *holder = &flag;
}

inline
Watchdog::Handler_Function::Handler_Function(void (*f)())
  : function(f) {
}

inline void
Watchdog::Handler_Function::act() const {
  (*function)();
}

inline
Watchdog::Pending_Element::Pending_Element(const Time& d,
                                           const Handler* h,
                                           bool* p)
  : deadline(d), handler(h), p_expired_flag(p) {
}

inline
Watchdog::Time::Time()
  : secs(0), microsecs(0) {
}

inline
Watchdog::Time::Time(unsigned int units)

  : secs(units / 100), microsecs((units * 10000) % 1000000) {
}

inline void
Watchdog::Time::set(int s, int m) {
    secs = s;
    microsecs = m;
  }

inline void
Watchdog::Time::reset() {
  secs = microsecs = 0;
}

inline int
Watchdog::Time::seconds() const {
  return secs;
}

inline int
Watchdog::Time::microseconds() const {
  return microsecs;
}

inline Watchdog::Time&
Watchdog::Time::operator+=(const Time& y) {
  int r_secs = secs + y.secs;
  int r_microsecs = microsecs + y.microsecs;
  if (r_microsecs >= 1000000) {
    ++r_secs;
    r_microsecs %= 1000000;
  }
  secs = r_secs;
  microsecs = r_secs;
  return *this;
}

inline Watchdog::Time&
Watchdog::Time::operator-=(const Time& y) {
  int r_secs = secs - y.secs;
  int r_microsecs = microsecs - y.microsecs;
  if (r_microsecs < 0) {
    --r_secs;
    r_microsecs += 1000000;
  }
  if (r_secs < 0)
    r_secs = r_microsecs = 0;
  secs = r_secs;
  microsecs = r_microsecs;
  return *this;
}

inline Watchdog::Time
Watchdog::Time::operator+(const Time& y) const {
  Time z = *this;
  z += y;
  return z;
}

inline Watchdog::Time
Watchdog::Time::operator-(const Time& y) const {
  Time z = *this;
  z -= y;
  return z;
}

inline bool
Watchdog::Time::operator==(const Time& y) const {
  return seconds() == y.seconds() && microseconds() == y.microseconds();
}

inline bool
Watchdog::Time::operator!=(const Time& y) const {
  return !(*this == y);
}

inline bool
Watchdog::Time::operator<(const Time& y) const {
  return seconds() < y.seconds()
    || (seconds() == y.seconds() && microseconds() < y.microseconds());
}

inline bool
Watchdog::Time::operator<=(const Time& y) const {
  return *this < y || *this == y;
}

inline bool
Watchdog::Time::operator>(const Time& y) const {
  return y < *this;
}

inline bool
Watchdog::Time::operator>=(const Time& y) const {
  return y <= *this;
}

inline void
Watchdog::reschedule() {
  set_timer(reschedule_time);
}

template <typename Flag_Base, typename Flag>
Watchdog::Watchdog(int units, const Flag_Base* volatile* holder, Flag& flag)
  : expired(false),
    handler(new Handler_Flag<Flag_Base, Flag>(holder, flag)) {
  if (units <= 0)
    throw std::invalid_argument("Watchdog constructor called with a"
                                " non-positive number of time units");
  in_critical_section = true;
  pending_position = new_watchdog_event(units, handler, &expired);
  in_critical_section = false;
}

inline
Watchdog::Watchdog(int units, void (*function)())
  : expired(false), handler(new Handler_Function(function)) {
  if (units <= 0)
    throw std::invalid_argument("Watchdog constructor called with a"
                                " non-positive number of time units");
  in_critical_section = true;
  pending_position = new_watchdog_event(units, handler, &expired);
  in_critical_section = false;
}

inline
Init::Init() {

  if (count++ == 0) {

    Watchdog::initialize();
  }
}

inline
Init::~Init() {

  if (--count == 0) {

    Watchdog::finalize();
  }
}

}

#ifdef PWL_SAVE_NDEBUG
#define NDEBUG 1
#undef PWL_SAVE_NDEBUG
#endif

#undef PACKAGE
#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION
#undef VERSION

#endif
